#version 330
#extension GL_ARB_separate_shader_objects : enable
precision highp float;
precision highp sampler2D;
precision highp sampler2DMS;
precision highp samplerCube;

layout(location = 0) smooth in vec2 _fin_fragTexCoord;

struct SFragmentInput
{
	vec2 fragTexCoord;
};

layout(location = 0) out vec4 _fout_outColor;

struct SFragmentOutput
{
	vec4 outColor;
};

 uniform sampler2D DepthSampler;
 uniform sampler2D NormalRoughMetalSampler;
 uniform sampler2D DiffuseSampler;
layout( std140) uniform ULightInfo
{
	mat4 InvViewProj;
	mat4 InvView;
	vec4 ViewDirection;
	vec4 ViewPosition;
	vec4 Position;
	vec4 Color;
	vec4 Ambient;
	float Radius;
} LightInfo;
 uniform samplerCube EnvironmentMapSampler;
 uniform sampler2D BRDFLUTSampler;



vec3 DepthToWorldPos(float pDepthValue, vec2 pUv, mat4 pInvViewProj)
{
	vec2 clipSpace = pUv * 2.0 - 1.0;
	vec4 projPos = vec4(clipSpace, pDepthValue, 1.0);
	projPos =  ((pInvViewProj) * (projPos)) ;
	return projPos.xyz / projPos.w;
}

vec3 UnpackNormalSpheremap(vec2 normal)
{
    vec4 nn = vec4(normal * 2.0 - 1.0, 1., -1.);
    float l = dot(nn.xyz, -nn.xyw);
    nn.z = l;
    nn.xy *= sqrt(l);
    return nn.xyz * 2. + vec3(0., 0., -1.);
}

vec3 UnpackNormalStd(vec3 normal)
{
	return normal * 2.0 - 1.0;
}

float getColorLuminance(vec3 color)
{
	return 0.299 * color.x + 0.587 * color.y + 0.114 * color.z;
}








float normalDistFuncGGX(float cosLh, float roughness)
{
	float alpha = roughness * roughness;
	float alphaSq = alpha * alpha;

	float denom = (cosLh * cosLh) * (alphaSq - 1.0f) + 1.0f;
	return alphaSq / ( 3.141592 * denom * denom);
}

float schlickGGX(float cosLi, float cosLo, float roughness)
{
	float r = roughness + 1.0f;
	float k = (r * r) / 8.0f;

    float t1 = cosLi / (cosLi * (1.0f - k) + k);
    float t2 = cosLo / (cosLo * (1.0f - k) + k);

	return t1 * t2;
}

vec3 fresnelSchlick(vec3 surfaceMetalColor, float cosTheta)
{
	return surfaceMetalColor + (vec3(1.0f, 1.0f, 1.0f) - surfaceMetalColor) * pow(1.0f - cosTheta, 5.0f);
}

vec3 computeBRDF(vec3 surfToLight, vec3 surfToView, vec3 surfaceNormal, float roughness, float metalness, vec3 surfaceColor  )
{
	vec3 halfVec = normalize(surfToLight + surfToView);

	float NoL = max(0.0f, dot(surfToLight, surfaceNormal));
	float specIntensity = max(0.0f, dot(halfVec, surfaceNormal));
	float NoV = max( 0.00001 , dot(surfToView, surfaceNormal));

	vec3 surfaceMetalColor = mix( vec3(0.04, 0.04, 0.04) , surfaceColor, metalness);

	vec3 F = fresnelSchlick(surfaceMetalColor, max(0.0f, dot(halfVec, surfToView)));
	float D = normalDistFuncGGX(specIntensity, roughness);
	float G = schlickGGX(NoL, NoV, roughness);

	vec3 diffuseBRDF = mix(vec3(1.0f, 1.0f, 1.0f) - F, vec3(0.0f, 0.0f, 0.0f), metalness) * surfaceColor;
	vec3 specularBRDF = (F * D * G) / max( 0.00001 , 4.0f * NoL * NoV);

	return (diffuseBRDF + specularBRDF) * NoL;
}

vec3 fresnelSchlickRoughness(float cosTheta, vec3 F0, float roughness)
{
    return F0 + (max(vec3(1.0f - roughness, 1.0f - roughness, 1.0f - roughness), F0) - F0) * pow(1.0f - cosTheta, 5.0f);
}

vec3 computeAmbientBRDF(vec3 surfToView, vec3 surfaceNormal, float roughness, float metalness, vec3 surfaceColor  )
{

	const float kCubemapMipRange = 7.f;

	vec3 surfaceMetalColor = mix( vec3(0.04, 0.04, 0.04) , surfaceColor, metalness);
	vec3 kS = fresnelSchlickRoughness(max(dot(surfaceNormal, surfToView), 0.0), surfaceMetalColor, roughness);
	vec3 kD = 1.0f - kS;
	kD *= 1.0f - metalness;


	vec3 irradiance =  textureLod(EnvironmentMapSampler, surfaceNormal, kCubemapMipRange) .rgb;
	vec3 diffuse = irradiance * surfaceColor;


	vec3 prefilteredColor =  textureLod(EnvironmentMapSampler, reflect(-surfToView, surfaceNormal), roughness * kCubemapMipRange) .rgb;
	vec2 envBRDF =  texture(BRDFLUTSampler, vec2(max(dot(surfaceNormal, surfToView), 0.0), roughness)) .rg;
	vec3 specular = prefilteredColor * (kS * envBRDF.x + envBRDF.y);

	vec3 ambient = kD * diffuse + specular;

	return ambient;
}

void FragmentMain( in SFragmentInput  fInput,  out SFragmentOutput  fOutput  )
{
	vec3 lightDirection =  LightInfo.Position .xyz;
	vec3 lightColor =  LightInfo.Color .xyz;
	vec3 ambientColor =  LightInfo.Ambient .rgb;
	mat4 invViewProj =  LightInfo.InvViewProj ;
	mat4 invView =  LightInfo.InvView ;
	vec3 viewPosition =  LightInfo.ViewPosition .xyz;

	vec3 surfaceColor =  texture(DiffuseSampler, fInput.fragTexCoord) .rgb;
	float depthValue =  texture(DepthSampler, fInput.fragTexCoord) .r;
	vec3 surfacePosition = DepthToWorldPos(depthValue, fInput.fragTexCoord, invViewProj);
	vec4 packedNormalRoughMetal =  texture(NormalRoughMetalSampler, fInput.fragTexCoord) ;
	vec3 viewSpaceNormal = normalize(UnpackNormalSpheremap(packedNormalRoughMetal.xy));
	vec3 surfaceNormal =  ((invView) * (vec4(viewSpaceNormal, 0.0f))) .xyz;
	float surfaceRoughness = max(0.01f, packedNormalRoughMetal.z);
	float surfaceMetalness = max(0.01f, packedNormalRoughMetal.w);

	vec3 surfToLightDir = normalize(-lightDirection);
	vec3 surfToView = viewPosition - surfacePosition;
	vec3 surfToViewDir = normalize(surfToView);

	vec3 lightDiffuse = computeBRDF( surfToLightDir,
										surfToViewDir,
										surfaceNormal,
										surfaceRoughness,
										surfaceMetalness,
										surfaceColor
										);

	vec3 lightAmbient = computeAmbientBRDF( surfToViewDir,
												surfaceNormal,
												surfaceRoughness,
												surfaceMetalness,
												surfaceColor
												);

	fOutput.outColor = vec4( lightDiffuse * lightColor + lightAmbient * ambientColor, 0.0f);
}


void main()
{
	SFragmentInput fInput;
	SFragmentOutput fOutput;

	fInput.fragTexCoord = _fin_fragTexCoord;
	FragmentMain(fInput, fOutput);
	_fout_outColor = fOutput.outColor;
}
